from functional_parsing_library.combinators.sequence.many import many
from functional_parsing_library.parser import Parser, U, S


def many_till(parser: Parser[U], until: Parser[S]) -> Parser[list[U]]:
    """
    Parses one or more matches for parser, followed by a match for until. Consumes the match for until.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = many_till(char('a'), char('b'))
    >>> parser('aab').result
    ['a', 'a']
    >>> parser('aab').remainder
    ''
    """
    return many(parser) < until



from functional_parsing_library.parser import Parser, S, U, ParseResults, CouldNotParse


def lookahead(parser: Parser[S], look_for: Parser[U]) -> Parser[S]:
    """
    Matches the same as the first parser, as long as it is followed by something matching the second parser. Does not
    consume whatever the second parser matches. For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = lookahead(char('a'), char('b'))
    >>> parser('abbb').result
    'a'
    >>> parser('abbb').remainder
    'bbb'
    >>> isinstance(parser('a'), CouldNotParse)
    True
    """
    def parser_(to_parse: str) -> ParseResults[S] | CouldNotParse:
        result_combined = (parser < look_for)(to_parse)
        if isinstance(result_combined, CouldNotParse):
            return result_combined
        return parser(to_parse)

    return Parser(parser_)



from functional_parsing_library.combinators.lookahead import lookahead
from functional_parsing_library.combinators.sequence.many import some
from functional_parsing_library.parser import Parser, U, S


def some_till_exclusive(parser: Parser[U], until: Parser[S]) -> Parser[list[U]]:
    """
    Parses zero or more matches for parser, followed by a match for until. Does not consume the match for until.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = some_till_exclusive(char('a'), char('b'))
    >>> parser('aab').result
    ['a', 'a']
    >>> parser('aab').remainder
    'b'
    >>> parser('b').result
    []
    """
    return lookahead(some(parser), look_for=until)
